@@ -4,13 +4,20 @@ use 5.006;
 package HTML::Restrict;
 
 use version;
-our $VERSION = 'v2.3.0';
+our $VERSION = 'v2.4.0';
 
 use Carp qw( croak );
 use Data::Dump qw( dump );
-use HTML::Parser ();
 use HTML::Entities qw( encode_entities );
-use Types::Standard 1.000001 qw[ Bool HashRef ArrayRef CodeRef ];
+use HTML::Parser ();
+use HTML::Restrict::Types qw(
+    ArrayRef
+    Bool
+    CodeRef
+    HashRef
+    Int
+    MaxParserLoops
+);
 use List::Util 1.33 qw( any none );
 use Scalar::Util qw( reftype weaken );
 use Sub::Quote 'quote_sub';
@@ -19,31 +26,37 @@ use URI ();
 use Moo 1.002000;
 use namespace::clean;
 
-has 'allow_comments' => (
+has allow_comments => (
     is      => 'rw',
     isa     => Bool,
     default => 0,
 );
 
-has 'allow_declaration' => (
+has allow_declaration => (
     is      => 'rw',
     isa     => Bool,
     default => 0,
 );
 
-has 'debug' => (
+has debug => (
     is      => 'rw',
     isa     => Bool,
     default => 0,
 );
 
-has 'parser' => (
+has max_parser_loops => (
+    is      => 'rw',
+    isa     => MaxParserLoops,
+    default => 25,
+);
+
+has parser => (
     is      => 'ro',
     lazy    => 1,
     builder => '_build_parser',
 );
 
-has 'rules' => (
+has rules => (
     is       => 'rw',
     isa      => HashRef,
     required => 0,
@@ -53,25 +66,25 @@ has 'rules' => (
     writer   => 'set_rules',
 );
 
-has 'strip_enclosed_content' => (
+has strip_enclosed_content => (
     is      => 'rw',
     isa     => ArrayRef,
     default => sub { [ 'script', 'style' ] },
 );
 
-has 'replace_img' => (
+has replace_img => (
     is      => 'rw',
     isa     => Bool | CodeRef,
     default => 0,
 );
 
-has 'trim' => (
+has trim => (
     is      => 'rw',
     isa     => Bool,
     default => 1,
 );
 
-has 'uri_schemes' => (
+has uri_schemes => (
     is       => 'rw',
     isa      => ArrayRef,
     required => 0,
@@ -80,7 +93,7 @@ has 'uri_schemes' => (
     writer   => 'set_uri_schemes',
 );
 
-has '_processed' => (
+has _processed => (
     is  => 'rw',
     isa => quote_sub(
         q{
@@ -91,14 +104,13 @@ has '_processed' => (
     clearer => '_clear_processed',
 );
 
-has '_stripper_stack' => (
+has _stripper_stack => (
     is      => 'rw',
     isa     => ArrayRef,
     default => sub { [] },
 );
 
 sub _build_parser {
-
     my $self  = shift;
     my $rules = shift;
 
@@ -276,11 +288,32 @@ sub _build_parser {
         ],
 
     );
-
 }
 
 sub process {
+    my $self = shift;
 
+    my $cleaned = $self->_process(@_);
+    return $cleaned if !$cleaned;
+
+    my $previous_iteration = $cleaned;
+
+    my $i = 1;    # We already cleaned once just above
+
+    while ( $i < $self->max_parser_loops ) {
+        $i++;
+        my $new = $self->_process($previous_iteration);
+        last if $new eq $previous_iteration;
+        if ( $i == $self->max_parser_loops ) {
+            die sprintf( 'Could not clean input after %s attempts', $i );
+        }
+
+        $previous_iteration = $new;
+    }
+    return $previous_iteration;
+}
+
+sub _process {
     my $self = shift;
 
     # returns undef if no value was passed
@@ -308,14 +341,12 @@ sub process {
     $self->_stripper_stack( [] );
 
     return $self->_processed;
-
 }
 
 # strip_enclosed_content tags could be nested in the source HTML, so we
 # maintain a stack of these tags.
 
 sub _delete_tag_from_stack {
-
     my $self        = shift;
     my $closing_tag = shift;
 
@@ -351,7 +382,7 @@ HTML::Restrict - Strip unwanted HTML tags and attributes
 
 =head1 VERSION
 
-version v2.3.0
+version v2.4.0
 
 =head1 SYNOPSIS
 
@@ -365,9 +396,7 @@ version v2.3.0
     # $processed now equals: 'i am bold'
 
     # Now, a less restrictive example:
-    use HTML::Restrict;
-
-    my $hr = HTML::Restrict->new(
+    $hr = HTML::Restrict->new(
         rules => {
             b   => [],
             img => [qw( src alt / )]
@@ -375,7 +404,7 @@ version v2.3.0
     );
 
     my $html = q[<body><b>hello</b> <img src="pic.jpg" alt="me" id="test" /></body>];
-    my $processed = $hr->process( $html );
+    $processed = $hr->process( $html );
 
     # $processed now equals: <b>hello</b> <img src="pic.jpg" alt="me" />
 
@@ -584,6 +613,30 @@ feature is off by default.
     $html = $hr->process( $html );
     # $html is now: "<!-- comments! -->foo"
 
+=item * max_parser_loops => [Integer]
+
+Defaults to 25.  Should never be less than 2.
+
+As of v2.4.0, calling C<process()> will force the parser to clean the text
+multiple times, stopping only once the text is no longer changed or once
+C<max_parser_loops> has been reached.
+
+The reason for this is that L<HTML::Parser> could take malformed HTML and turn
+it into well formed HTML.  This can defeat our processing logic and allow
+malicious input to be returned.  In order to mitigate this, we will clean all
+input at least two times.  If the second attempt at cleaning does not match
+the previous attempt, we will make a third attempt and so on.  This helps to
+ensure that we get the expected output.
+
+If we are unable to get unchanged values after reaching C<max_parser_loops>, an
+exception will be thrown.  Returning partially cleaned text would be wrong, as
+would be returning C<undef> or an empty string.  Throwing an exception forces
+the user to choose the appropriate way of dealing with this.
+
+If you choose to set this value, please note that it can be no less than 2, or
+the parser will never be able to make a comparison with a previous value.  An
+exception will be thrown if you attempt to set this to a value less than 2.
+
 =item * replace_img => [0|1|CodeRef]
 
 Set the value to true if you'd like to have img tags replaced with
